/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Cmad: areaDetector "driver" for CameraMan enabling CameraMan to capture
 * frames from an areaDetector IOC.
 *
 * We do not write anything to client->err_msg from callbacks nor other
 * internal functions; only public API functions set client->err_msg if they
 * return an error status.
 *
 * The CA client library serializes all callbacks within a CA context, and
 * this library uses only one CA context, so only one callback will ever be
 * executing at a time.
 */

#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <cadef.h>
#include <epicsEvent.h>
#include <epicsMutex.h>

#define epicsExportSharedSymbols
#include "cmad.h"
#include "cmad_internal.h"

#define CMAD_SEQ_NUM_MAX      2047
#define CMAD_SEQ_NUM_SIZE     CMAD_SEQ_NUM_MAX + 1
#define CMAD_SEQ_NUM_WIN_SIZE 120

/* areaDetector NDColorModeRGB1 enum constant from NDColorMode_t enum */
#define CMAD_AD_COLOR_MODE_RGB1 2

#define CMAD_NUM_IMAGE_PIXEL_COMPONENTS 3

#define CMAD_NUM_CHANNELS 6

#define CMAD_CH_NX         0
#define CMAD_CH_NY         1
#define CMAD_CH_NZ         2
#define CMAD_CH_COLOR_MODE 3
#define CMAD_CH_IMAGE      4
#define CMAD_CH_IMAGE_ID   5

#define CMAD_PV_SUFFIX_NX         "ArraySize0_RBV"
#define CMAD_PV_SUFFIX_NY         "ArraySize1_RBV"
#define CMAD_PV_SUFFIX_NZ         "ArraySize2_RBV"
#define CMAD_PV_SUFFIX_COLOR_MODE "ColorMode_RBV"
#define CMAD_PV_SUFFIX_IMAGE      "ArrayData"
#define CMAD_PV_SUFFIX_IMAGE_ID   "UniqueId_RBV"

struct cmad_client {
  const char *pv_prefix;
  unsigned long image_width;
  unsigned long image_height;
  unsigned long image_size;
  unsigned long image_size_in_bytes;
  double timeout;
  epicsMutexId mutex_id;
  char *err_msg;
  int ch_connected[CMAD_NUM_CHANNELS];
  int connected;
  int ch_ready[CMAD_NUM_CHANNELS];
  int ready;
  int ready_except_image;
  int capture_enabled;
  unsigned char *image;
  ptrdiff_t image_seq;
  epicsEventId image_seq_eid;
  chid ch_nx;
  chid ch_ny;
  chid ch_nz;
  chid ch_color_mode;
  chid ch_image;
  chid ch_image_id;
};

static epicsThreadOnceId cmad_once_flag = EPICS_THREAD_ONCE_INIT;
static epicsMutexId cmad_mutex_id;
static struct ca_client_context *cmad_ca_context = NULL;
static unsigned long cmad_client_ref_count = 0;
static const char cmad_seq_num_dummy[CMAD_SEQ_NUM_SIZE];

static void cmad_init(void);
static void cmad_init_once(void *);
static int cmad_attach_to_ca_context_locked(void);
static void cmad_detach_from_ca_context_locked(void);
static int cmad_increment_client_ref_count_locked(void);
static int cmad_decrement_client_ref_count_locked(void);
static int cmad_ca_start_up_attached(struct cmad_client *);
static void cmad_ca_shut_down_attached(struct cmad_client *);
static const char *cmad_pv_name_new(const char *, ...);
static int cmad_set_error(struct cmad_client *, const char *, ...);
static int cmad_set_error_locked(struct cmad_client *, const char *, ...);
static int cmad_wait_connected_ready_locked(struct cmad_client *, double);
static void cmad_update_connected_locked(struct cmad_client *);
static void cmad_update_ready_locked(struct cmad_client *);
static ptrdiff_t cmad_increment_sequence_number(ptrdiff_t);
static int cmad_is_sequence_number_valid(ptrdiff_t, ptrdiff_t);
static void *cmad_encode_sequence_number(ptrdiff_t);
static ptrdiff_t cmad_decode_sequence_number(void *);
static void cmad_rgb24_to_bgr24(unsigned char *, unsigned long);

static void cmad_on_exception(struct exception_handler_args);
static void cmad_on_nx_connection_state_change(
    struct connection_handler_args);
static void cmad_on_ny_connection_state_change(
    struct connection_handler_args);
static void cmad_on_nz_connection_state_change(
    struct connection_handler_args);
static void cmad_on_color_mode_connection_state_change(
    struct connection_handler_args);
static void cmad_on_image_connection_state_change(
    struct connection_handler_args);
static void cmad_on_image_id_connection_state_change(
    struct connection_handler_args);
static void cmad_on_connection_state_change(
    struct connection_handler_args, int);
static void cmad_on_nx_state_change(struct event_handler_args);
static void cmad_on_ny_state_change(struct event_handler_args);
static void cmad_on_nz_state_change(struct event_handler_args);
static void cmad_on_color_mode_state_change(struct event_handler_args);
static void cmad_on_image_id_state_change(struct event_handler_args);
static void cmad_on_image_get(struct event_handler_args);

struct cmad_client *
cmad_client_new(const char *pv_prefix, unsigned long image_width,
    unsigned long image_height, double timeout)
{
  epicsMutexLockStatus mutex_status;
  int locked;
  struct cmad_client *client;
  int attached;
  int status;
  size_t i;

  locked = 0;
  client = NULL;
  attached = 0;

  cmad_init();

  mutex_status = epicsMutexLock(cmad_mutex_id);
  if (mutex_status != epicsMutexLockOK) goto err;
  locked = 1;

  status = cmad_attach_to_ca_context_locked();
  if (status == -1) goto err;
  attached = 1;

  client = calloc(1, sizeof(*client));
  if (client == NULL) goto err;

  client->pv_prefix = NULL;
  client->err_msg = NULL;
  client->image = NULL;

  if (!pv_prefix) goto err;
  client->pv_prefix = strdup(pv_prefix);
  if (client->pv_prefix == NULL) goto err;
  if (image_width == 0) goto err;
  client->image_width = image_width;
  if (image_height == 0) goto err;
  client->image_height = image_height;
  if ((ULONG_MAX / image_width) < image_height) goto err;
  client->image_size = image_width * image_height;
  if ((ULONG_MAX / client->image_size) < CMAD_NUM_IMAGE_PIXEL_COMPONENTS) {
      goto err;
  }
  client->image_size = client->image_size * CMAD_NUM_IMAGE_PIXEL_COMPONENTS;
  if ((ULONG_MAX / client->image_size) < sizeof(*client->image)) {
      goto err;
  }
  client->image_size_in_bytes = client->image_size * sizeof(*client->image);
  if (timeout <= 0.0) goto err;
  client->timeout = timeout;
  client->mutex_id = epicsMutexCreate();
  if (!client->mutex_id) goto err;
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) client->ch_connected[i] = 0;
  client->connected = 0;
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) client->ch_ready[i] = 0;
  client->ready = 0;
  client->ready_except_image = 0;
  client->capture_enabled = 0;
  client->image = calloc(client->image_size, sizeof(*client->image));
  if (client->image == NULL) goto err;
  client->image_seq = 0;
  client->image_seq_eid = epicsEventCreate(epicsEventEmpty);
  if (!client->image_seq_eid) goto err;

  status = cmad_ca_start_up_attached(client);
  if (status == -1) goto err;

  status = cmad_increment_client_ref_count_locked();
  if (status == -1) goto err;

  goto out;

err:
  if (client != NULL) {
    if (client->pv_prefix != NULL) free((char *)client->pv_prefix);
    if (client->mutex_id) epicsMutexDestroy(client->mutex_id);
    if (client->image != NULL) free(client->image);
    if (client->image_seq_eid) epicsEventDestroy(client->image_seq_eid);
    free(client);
    client = NULL;
  }
  if (attached) cmad_detach_from_ca_context_locked();

out:
  if (locked) epicsMutexUnlock(cmad_mutex_id);
  return client;
}

void
cmad_client_free(struct cmad_client *client)
{
  int status;

  if (client == NULL) return;

  cmad_init();

  epicsMutexMustLock(cmad_mutex_id);

  status = cmad_attach_to_ca_context_locked();
  if (status == -1) goto err;

  cmad_ca_shut_down_attached(client);

err:
  free((char *)client->pv_prefix);
  if (client->err_msg != NULL) free(client->err_msg);
  epicsMutexDestroy(client->mutex_id);
  free(client->image);
  epicsEventDestroy(client->image_seq_eid);
  free(client);

  (void)cmad_decrement_client_ref_count_locked();

  cmad_detach_from_ca_context_locked();

  epicsMutexUnlock(cmad_mutex_id);
}

const char *
cmad_error(struct cmad_client *client)
{
  const char *rv;
  epicsMutexLockStatus mutex_status;

  rv = NULL;

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return rv;

  rv = client->err_msg;

  epicsMutexUnlock(client->mutex_id);

  return rv;
}

int
cmad_is_connected(struct cmad_client *client)
{
  int rv;
  epicsMutexLockStatus mutex_status;

  rv = 0;

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return rv;

  rv = client->connected ? 1 : 0;

  epicsMutexUnlock(client->mutex_id);

  return rv;
}

int
cmad_is_ready(struct cmad_client *client)
{
  int rv;
  epicsMutexLockStatus mutex_status;

  rv = 0;

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return rv;

  rv = client->ready ? 1 : 0;

  epicsMutexUnlock(client->mutex_id);

  return rv;
}

int
cmad_start_image_capture(struct cmad_client *client)
{
  epicsMutexMustLock(client->mutex_id);

  client->capture_enabled = 1;

  epicsMutexUnlock(client->mutex_id);

  return 0;
}

int
cmad_stop_image_capture(struct cmad_client *client)
{
  epicsMutexMustLock(client->mutex_id);

  client->capture_enabled = 0;

  epicsMutexUnlock(client->mutex_id);

  return 0;
}

int
cmad_capture_image(struct cmad_client *client, unsigned char *buf,
    size_t buf_size)
{
  int rv;
  int locked;
  double timeout;
  epicsEventId image_seq_eid;
  epicsEventWaitStatus event_status;
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  rv = -1;
  locked = 0;

  epicsMutexMustLock(client->mutex_id);
  locked = 1;

  if (buf_size < client->image_size) {
    cmad_set_error_locked(client, "buffer size is too small");
    goto err;
  }

  client->image_seq = cmad_increment_sequence_number(client->image_seq);

  event_status = epicsEventTryWait(client->image_seq_eid);
  if (event_status == epicsEventWaitError) {
    cmad_set_error_locked(client, "unable to clear image-seq signal");
    goto err;
  }

  timeout = client->timeout;
  image_seq_eid = client->image_seq_eid;

  epicsMutexUnlock(client->mutex_id);
  locked = 0;

  status = epicsTimeGetCurrent(&start_time);
  if (status == -1) {
    cmad_set_error(client, "unable to obtain current time");
    goto err;
  }

  event_status = epicsEventWaitWithTimeout(image_seq_eid, timeout);
  if (event_status != epicsEventWaitOK) {
    cmad_set_error(client,
        (event_status == epicsEventWaitTimeout)
            ? "timed out waiting for image"
            : "unable to wait for image-seq signal");
    goto err;
  }

  status = epicsTimeGetCurrent(&now_time);
  if (status == -1) {
    cmad_set_error(client, "unable to obtain current time");
    goto err;
  }

  elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
  if (elapsed_time < 0.0) {
    /* Time advanced backward; consider now to be the new start time, and
     * consider no time to have elapsed. */
    start_time = now_time;
    elapsed_time = 0.0;
  } else if (elapsed_time > timeout) {
    cmad_set_error(client, "timed out waiting for image");
    goto err;
  }

  epicsMutexMustLock(client->mutex_id);
  locked = 1;

  status = cmad_wait_connected_ready_locked(client, timeout - elapsed_time);
  if (status == -1) {
    cmad_set_error_locked(client, "timed out waiting for image");
    goto err;
  }

  (void)memcpy(buf, client->image, client->image_size_in_bytes);

  rv = 0;

err:
  if (locked) epicsMutexUnlock(client->mutex_id);

  return rv;
}

int
cmad_capture_image_quick(struct cmad_client *client, unsigned char *buf,
    size_t buf_size)
{
  int rv;
  int status;

  rv = -1;

  epicsMutexMustLock(client->mutex_id);

  if (buf_size < client->image_size) {
    cmad_set_error_locked(client, "buffer size is too small");
    goto err;
  }

  status = cmad_wait_connected_ready_locked(client, client->timeout);
  if (status == -1) {
    cmad_set_error_locked(client,
        "errored or timed out waiting for connected and ready");
    goto err;
  }

  (void)memcpy(buf, client->image, client->image_size_in_bytes);

  rv = 0;

err:
  epicsMutexUnlock(client->mutex_id);

  return rv;
}

/*
 * Set the client's timeout in seconds.  The timeout must be greater than 0.
 * Return 0 on success and -1 on error.
 */
int
cmad_set_timeout(struct cmad_client *client, double timeout)
{
  int rv;

  rv = -1;

  if (timeout <= 0.0) {
    cmad_set_error(client, "timeout is less than or equal to 0");
    goto err;
  }

  epicsMutexMustLock(client->mutex_id);

  client->timeout = timeout;

  epicsMutexUnlock(client->mutex_id);

  rv = 0;

err:
  return rv;
}

/*
 * Initialize the library.  It is permissible to call this function more than
 * once.
 */
static void
cmad_init(void)
{
  epicsThreadOnce(&cmad_once_flag, cmad_init_once, NULL);
}

/*
 * Called by epicsThreadOnce() exactly once to initialize the library.
 */
static void
cmad_init_once(void *arg)
{
  cmad_mutex_id = epicsMutexMustCreate();
}

/*
 * Attach the current thread to the EPICS CA context.  The caller must already
 * hold the Cmad mutex.  If the context does not exist, create it.  Return
 * 0 on success and -1 on error.  If this function returns 0 (success),
 * the caller is responsible for detaching from the context by calling
 * cmad_detach_from_ca_context_locked().
 */
static int
cmad_attach_to_ca_context_locked(void)
{
  int rv;
  int status;
  int created_ca_context;

  rv = -1;
  created_ca_context = 0;

  if (cmad_ca_context) {
    status = ca_attach_context(cmad_ca_context);
    if (!(status == ECA_NORMAL || status == ECA_ISATTACHED)) goto err;
  } else {
    status = ca_context_create(ca_enable_preemptive_callback);
    if (status != ECA_NORMAL) goto err;
    created_ca_context = 1;
    cmad_ca_context = ca_current_context();
    if (!cmad_ca_context) goto err;
    status = ca_add_exception_event(cmad_on_exception, NULL);
    if (status != ECA_NORMAL) goto err;
  }

  rv = 0;
  goto out;

err:
  if (created_ca_context) {
    ca_context_destroy();
    cmad_ca_context = NULL;
  }

out:
  return rv;
}

/*
 * Detach the current thread from the EPICS CA context.  The caller must
 * already hold the Cmad mutex.  If the client reference count is zero, destroy
 * the context.
 */
static void
cmad_detach_from_ca_context_locked(void)
{
  int status;

  status = cmad_attach_to_ca_context_locked();
  if (status == -1) goto err;

  if (cmad_client_ref_count > 0) {
    ca_detach_context();
  } else {
    ca_context_destroy();
    cmad_ca_context = NULL;
  }

err:
  return;
}

/*
 * Increment the client reference count by one.  The caller must already hold
 * the Cmad mutex.  Return 0 on success and -1 on error.
 */
static int
cmad_increment_client_ref_count_locked(void)
{
  if (cmad_client_ref_count == ULONG_MAX) return -1;
  cmad_client_ref_count++;
  return 0;
}

/*
 * Decrement the client reference count by one.  The caller must already hold
 * the Cmad mutex.  Return 0 on success and -1 on error.
 */
static int
cmad_decrement_client_ref_count_locked(void)
{
  if (cmad_client_ref_count == 0) return -1;
  cmad_client_ref_count--;
  return 0;
}

/*
 * Create the client's channels and subscribe for change events.  The caller
 * must have already attached to the CA context.  The user private void pointer
 * variable (retrievable with ca_puser()) of each channel is set to the client.
 */
static int
cmad_ca_start_up_attached(struct cmad_client *client)
{
  int rv;
  epicsMutexLockStatus mutex_status;
  int i;
  const char *pv_suffix[CMAD_NUM_CHANNELS];
  const char *pv_name[CMAD_NUM_CHANNELS];
  chtype ch_type[CMAD_NUM_CHANNELS];
  unsigned long ch_count[CMAD_NUM_CHANNELS];
  void *ch_conn_callback[CMAD_NUM_CHANNELS];
  void *ch_subs_callback[CMAD_NUM_CHANNELS];
  chid *ch[CMAD_NUM_CHANNELS];
  int ch_created[CMAD_NUM_CHANNELS];
  int status;

  rv = -1;

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return rv;

  i = CMAD_CH_NX;
  pv_suffix[i] = CMAD_PV_SUFFIX_NX;
  pv_name[i] = NULL;
  ch_type[i] = DBR_LONG;
  ch_count[i] = 1;
  ch_conn_callback[i] = cmad_on_nx_connection_state_change;
  ch_subs_callback[i] = cmad_on_nx_state_change;
  ch[i] = &client->ch_nx;
  ch_created[i] = 0;

  i = CMAD_CH_NY;
  pv_suffix[i] = CMAD_PV_SUFFIX_NY;
  pv_name[i] = NULL;
  ch_type[i] = DBR_LONG;
  ch_count[i] = 1;
  ch_conn_callback[i] = cmad_on_ny_connection_state_change;
  ch_subs_callback[i] = cmad_on_ny_state_change;
  ch[i] = &client->ch_ny;
  ch_created[i] = 0;

  i = CMAD_CH_NZ;
  pv_suffix[i] = CMAD_PV_SUFFIX_NZ;
  pv_name[i] = NULL;
  ch_type[i] = DBR_LONG;
  ch_count[i] = 1;
  ch_conn_callback[i] = cmad_on_nz_connection_state_change;
  ch_subs_callback[i] = cmad_on_nz_state_change;
  ch[i] = &client->ch_nz;
  ch_created[i] = 0;

  i = CMAD_CH_COLOR_MODE;
  pv_suffix[i] = CMAD_PV_SUFFIX_COLOR_MODE;
  pv_name[i] = NULL;
  ch_type[i] = DBR_ENUM;
  ch_count[i] = 1;
  ch_conn_callback[i] = cmad_on_color_mode_connection_state_change;
  ch_subs_callback[i] = cmad_on_color_mode_state_change;
  ch[i] = &client->ch_color_mode;
  ch_created[i] = 0;

  i = CMAD_CH_IMAGE;
  pv_suffix[i] = CMAD_PV_SUFFIX_IMAGE;
  pv_name[i] = NULL;
  ch_type[i] = DBR_CHAR;
  ch_count[i] = 0;
  ch_conn_callback[i] = cmad_on_image_connection_state_change;
  ch_subs_callback[i] = NULL;
  ch[i] = &client->ch_image;
  ch_created[i] = 0;

  i = CMAD_CH_IMAGE_ID;
  pv_suffix[i] = CMAD_PV_SUFFIX_IMAGE_ID;
  pv_name[i] = NULL;
  ch_type[i] = DBR_LONG;
  ch_count[i] = 1;
  ch_conn_callback[i] = cmad_on_image_id_connection_state_change;
  ch_subs_callback[i] = cmad_on_image_id_state_change;
  ch[i] = &client->ch_image_id;
  ch_created[i] = 0;

  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    pv_name[i] = cmad_pv_name_new("%s%s", client->pv_prefix, pv_suffix[i]);
    if (pv_name[i] == NULL) goto err;

    status = ca_create_channel(pv_name[i], ch_conn_callback[i], client, 0,
        ch[i]);
    if (status != ECA_NORMAL) goto err;
    ch_created[i] = 1;

    if (ch_subs_callback[i] != NULL) {
      status = ca_create_subscription(ch_type[i], ch_count[i], *ch[i],
          DBE_VALUE, ch_subs_callback[i], NULL, NULL);
      if (status != ECA_NORMAL) goto err;
    }
  }

  status = ca_flush_io();
  if (status != ECA_NORMAL) goto err;

  rv = 0;
  goto out;

err:
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    if (ch_created[i]) (void)ca_clear_channel(*ch[i]);
  }
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    if (ch_created[i]) {
      (void)ca_flush_io();
      break;
    }
  }

out:
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    if (pv_name[i] != NULL) free((char *)pv_name[i]);
  }

  epicsMutexUnlock(client->mutex_id);

  return rv;
}

/*
 * Destroy the client's channels and subscriptions.  The caller must have
 * already attached to the CA context.
 */
static void
cmad_ca_shut_down_attached(struct cmad_client *client)
{
  (void)ca_clear_channel(client->ch_nx);
  (void)ca_clear_channel(client->ch_ny);
  (void)ca_clear_channel(client->ch_nz);
  (void)ca_clear_channel(client->ch_color_mode);
  (void)ca_clear_channel(client->ch_image);
  (void)ca_clear_channel(client->ch_image_id);
  (void)ca_flush_io();
}

/*
 * Create a new PV name string according to the specified printf-style format
 * and arguments.  Return a pointer to the new PV name string on success and
 * NULL on error.  If this function returns a pointer to a string (success),
 * the caller is responsible for freeing the string.
 */
static const char *
cmad_pv_name_new(const char *fmt, ...)
{
  va_list ap;
  char *buf;
  size_t buf_size;
  size_t tmp_buf_size;
  int i;
  int num_chars_printed;

  buf = NULL;
  buf_size = 128;

  /*
   * The best would be to use vasprintf, but not all platforms have it, and
   * it's not in EPICS Base libCom, so use a poor man's alternative.
   */
  for (i = 0; i < 2; i++) {
    buf = calloc(buf_size, sizeof(*buf));
    if (buf == NULL) goto err;
    va_start(ap, fmt);
    num_chars_printed = vsnprintf(buf, buf_size, fmt, ap);
    va_end(ap);
    if (num_chars_printed < 0) {
      goto err;
    } else if ((size_t)num_chars_printed >= buf_size) {
      tmp_buf_size = num_chars_printed;
      buf_size = tmp_buf_size + 1;
      if (buf_size < tmp_buf_size) goto err;
      free(buf);
      buf = NULL;
    } else {
      return buf;
    }
  }

err:
  if (buf != NULL) free(buf);
  return NULL;
}

/*
 * Set the client's error message according to the specified printf-style
 * format and arguments.  Return the number of characters written (not
 * including the trailing '\0') on success and -1 on error.
 */
static int
cmad_set_error(struct cmad_client *client, const char *fmt, ...)
{
  va_list ap;
  int rv;
  char *buf;
  size_t buf_size;
  size_t tmp_buf_size;
  int i;
  int num_chars_printed;

  rv = -1;

  epicsMutexMustLock(client->mutex_id);

  if (client->err_msg != NULL) {
    free(client->err_msg);
    client->err_msg = NULL;
  }

  buf = NULL;
  buf_size = 128;

  /*
   * The best would be to use vasprintf, but not all platforms have it, and
   * it's not in EPICS Base libCom, so use a poor man's alternative.
   */
  for (i = 0; i < 2; i++) {
    buf = calloc(buf_size, sizeof(*buf));
    if (buf == NULL) goto err;
    va_start(ap, fmt);
    num_chars_printed = vsnprintf(buf, buf_size, fmt, ap);
    va_end(ap);
    if (num_chars_printed < 0) {
      goto err;
    } else if ((size_t)num_chars_printed >= buf_size) {
      tmp_buf_size = num_chars_printed;
      buf_size = tmp_buf_size + 1;
      if (buf_size < tmp_buf_size) goto err;
      free(buf);
      buf = NULL;
    } else {
      client->err_msg = buf;
      break;
    }
  }

  rv = num_chars_printed;
  goto out;

err:
  if (buf != NULL) free(buf);

out:
  epicsMutexUnlock(client->mutex_id);

  return rv;
}

/*
 * Set the client's error message according to the specified printf-style
 * format and arguments.  The caller must already hold the client's mutex.
 * Return the number of characters written (not including the trailing '\0') on
 * success and -1 on error.
 */
static int
cmad_set_error_locked(struct cmad_client *client, const char *fmt, ...)
{
  va_list ap;
  int rv;
  char *buf;
  size_t buf_size;
  size_t tmp_buf_size;
  int i;
  int num_chars_printed;

  rv = -1;

  if (client->err_msg != NULL) {
    free(client->err_msg);
    client->err_msg = NULL;
  }

  buf = NULL;
  buf_size = 128;

  /*
   * The best would be to use vasprintf, but not all platforms have it, and
   * it's not in EPICS Base libCom, so use a poor man's alternative.
   */
  for (i = 0; i < 2; i++) {
    buf = calloc(buf_size, sizeof(*buf));
    if (buf == NULL) goto err;
    va_start(ap, fmt);
    num_chars_printed = vsnprintf(buf, buf_size, fmt, ap);
    va_end(ap);
    if (num_chars_printed < 0) {
      goto err;
    } else if ((size_t)num_chars_printed >= buf_size) {
      tmp_buf_size = num_chars_printed;
      buf_size = tmp_buf_size + 1;
      if (buf_size < tmp_buf_size) goto err;
      free(buf);
      buf = NULL;
    } else {
      client->err_msg = buf;
      break;
    }
  }

  rv = num_chars_printed;
  goto out;

err:
  if (buf != NULL) free(buf);

out:
  return rv;
}

/*
 * Wait for the client to be connected and ready, up to the specified timeout
 * in seconds.  The caller must already hold the client's mutex.  Return 0 on
 * success and -1 on error or timeout.
 */
int
cmad_wait_connected_ready_locked(struct cmad_client *client, double timeout)
{
  int rv;
  int locked;
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  rv = -1;
  locked = 1;

  /* Quickly check the common case where the client is connected and ready. */
  if (client->connected && client->ready) return 0;

  epicsMutexUnlock(client->mutex_id);
  locked = 0;

  status = epicsTimeGetCurrent(&start_time);
  if (status == -1) goto err;

  for (;;) {
    epicsMutexMustLock(client->mutex_id);
    locked = 1;

    if (client->connected && client->ready) break;

    epicsMutexUnlock(client->mutex_id);
    locked = 0;

    status = epicsTimeGetCurrent(&now_time);
    if (status == -1) goto err;

    elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
    if (elapsed_time < 0.0) {
      /* Time advanced backward; consider now to be the new start time. */
      start_time = now_time;
    } else if (elapsed_time > timeout) {
      goto err;
    }

    epicsThreadSleep(0.1);
  }

  rv = 0;

err:
  if (!locked) epicsMutexMustLock(client->mutex_id);

  return rv;
}

/*
 * Update the client's connected field based on its ch_connected array field.
 * The caller must already hold the client's mutex.
 */
static void
cmad_update_connected_locked(struct cmad_client *client)
{
  int i;

  client->connected = 1;
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    if (!client->ch_connected[i]) {
      client->connected = 0;
      break;
    }
  }
}

/*
 * Update the client's ready and ready_except_image fields based on its
 * ch_ready array field.  The caller must already hold the client's mutex.
 */
static void
cmad_update_ready_locked(struct cmad_client *client)
{
  int i;

  client->ready = 1;
  client->ready_except_image = 1;
  for (i = 0; i < CMAD_NUM_CHANNELS; i++) {
    if (!client->ch_ready[i]) {
      client->ready = 0;
      if (i != CMAD_CH_IMAGE) client->ready_except_image = 0;
    }
  }
}

/*
 * Increment the sequence number by one.  If incrementing the sequence number
 * by one would cause it to exceed the maximum sequence number, it is rolled
 * over to 0.  Return the incremented sequence number.
 */
static ptrdiff_t
cmad_increment_sequence_number(ptrdiff_t seq_num)
{
  return (seq_num < CMAD_SEQ_NUM_MAX) ? (seq_num + 1) : 0;
}

/*
 * Test whether seq_num is a valid sequence number.  It is valid if it's within
 * the window of size CMAD_SEQ_NUM_WIN_SIZE starting at earliest_seq_num.  If
 * the window would extend beyond the maximum sequence number, it wraps and
 * continues from the minimum sequence number (0).  Return 1 if valid and 0 if
 * invalid.
 */
static int
cmad_is_sequence_number_valid(ptrdiff_t seq_num, ptrdiff_t earliest_seq_num)
{
  ptrdiff_t avail_win_size_before_wrap;
  ptrdiff_t win_size_after_wrap;

  if (seq_num == earliest_seq_num) return 1;

  avail_win_size_before_wrap = CMAD_SEQ_NUM_SIZE - earliest_seq_num;
  if (avail_win_size_before_wrap >= CMAD_SEQ_NUM_WIN_SIZE) {
    return (seq_num >= earliest_seq_num &&
            seq_num < (earliest_seq_num + CMAD_SEQ_NUM_WIN_SIZE))
        ? 1
        : 0;
  } else {
    win_size_after_wrap = CMAD_SEQ_NUM_WIN_SIZE - avail_win_size_before_wrap;
    return ((seq_num >= earliest_seq_num && seq_num < CMAD_SEQ_NUM_SIZE) ||
            seq_num < win_size_after_wrap)
        ? 1
        : 0;
  }
}

/*
 * Encode the sequence number, seq_num, as a generic pointer.  The
 * generic pointer can be decoded back into the sequence number via the
 * cmad_decode_sequence_number() function.  Return the encoded generic pointer.
 *
 * This provides a way to avoid allocating memory for a sequence number for
 * every image-get and then deallocating it in each callback.  This is achieved
 * by effectively encoding the sequence number in a "void *" pointer in a
 * portable way.  The downside is that this requires one byte of memory for
 * every possible sequence number regardless of whether a given sequence number
 * is being used.  This won't scale well for a large sequence number range, but
 * for a small range, it's OK.
 *
 * A simpler implementation would just index into a static array of the same
 * type as the sequence number, initialized to every possible sequence number.
 * No pointer arithmetic would be required, and the sequence number could
 * have a more normal type (i.e. "unsigned short" instead of ptrdiff_t).  The
 * downside is that it would require more memory since the type would need to
 * be larger than char, the smallest C type (one byte), to represent more than
 * 256 sequence numbers.  The chosen implementation is able to use char for
 * the type of the static dummy array and uses pointer arithmetic to encode a
 * sequence number larger than 255 as a "void *" pointer.
 *
 * Obviously, another approach would be to just try the rapid allocations and
 * deallocations and see if it works.  Or, a memory pool could be tried.  These
 * would not require memory to hold all possible sequence numbers like the
 * chosen implementation, but they would add complexity and require the CA
 * client library to be bug-free and guarantee that it calls the callbacks in
 * all situations so that the memory can be freed.  The CA client library does
 * guarantee this, but I'm reluctant to trust it this much, so a solution that
 * avoids the situation altogether is appealing.
 */
static void *
cmad_encode_sequence_number(ptrdiff_t seq_num)
{
  /* Cast original type "const char *" to "void *". */
  return (void *)(cmad_seq_num_dummy + seq_num);
}

/*
 * Decode the generic pointer, seq_num, into a sequence number.  The
 * generic pointer must be a sequence number that has been encoded via the
 * cmad_encode_sequence_number() function.  Return the decoded sequence number.
 *
 * See the comment for the cmad_encode_sequence_number() function for more
 * details on the implementation choice.
 */
static ptrdiff_t
cmad_decode_sequence_number(void *seq_num)
{
  /* Cast "void *" to original type "const char *". */
  return (const char *)seq_num - cmad_seq_num_dummy;
}

/*
 * Convert an RGB24 image into a BGR24 image in-place.  Each pixel is
 * represented as three bytes of color (hence the 24 in the image format
 * names).  The fastest changing dimension is color, then width, and then
 * height (the slowest).
 */
static void
cmad_rgb24_to_bgr24(unsigned char *image, unsigned long image_size)
{
  unsigned long i;
  unsigned char tmp;

  for (i = 2; i < image_size; i += 3) {
    tmp = image[i];
    image[i] = image[i - 2];
    image[i - 2] = tmp;
  }
}

/*
 * Exception handler called when a CA exception occurs.  This library provides
 * no way to report exceptions to the user, so just drop them.  This is
 * preferred over the default which prints a message to the standard output
 * stream and terminates execution if the exception is severe.
 */
static void
cmad_on_exception(struct exception_handler_args args)
{
  /* NOOP */
}

/*
 * Callback called when the nx channel's connection state changes.
 */
static void
cmad_on_nx_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_NX);
}

/*
 * Callback called when the ny channel's connection state changes.
 */
static void
cmad_on_ny_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_NY);
}

/*
 * Callback called when the nz channel's connection state changes.
 */
static void
cmad_on_nz_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_NZ);
}

/*
 * Callback called when the color-mode channel's connection state changes.
 */
static void
cmad_on_color_mode_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_COLOR_MODE);
}

/*
 * Callback called when the image channel's connection state changes.
 */
static void
cmad_on_image_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_IMAGE);
}

/*
 * Callback called when the image-ID channel's connection state changes.
 */
static void
cmad_on_image_id_connection_state_change(struct connection_handler_args args)
{
  cmad_on_connection_state_change(args, CMAD_CH_IMAGE_ID);
}

/*
 * Callback called when a channel's connection state changes.
 */
static void
cmad_on_connection_state_change(struct connection_handler_args args,
    int ch_index)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  client->ch_connected[ch_index] = (args.op == CA_OP_CONN_UP) ? 1 : 0;

  cmad_update_connected_locked(client);

  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for nx CA subscription state change events.
 */
static void
cmad_on_nx_state_change(struct event_handler_args args)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;
  dbr_long_t nx;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_NX] = 0;
    goto out;
  }

  nx = *((dbr_long_t *)args.dbr);
  if (nx < 0) nx = 0;
  client->ch_ready[CMAD_CH_NX] =
     (nx == CMAD_NUM_IMAGE_PIXEL_COMPONENTS) ? 1 : 0;

out:
  cmad_update_ready_locked(client);

  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for ny CA subscription state change events.
 */
static void
cmad_on_ny_state_change(struct event_handler_args args)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;
  dbr_long_t ny;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_NY] = 0;
    goto out;
  }

  ny = *((dbr_long_t *)args.dbr);
  if (ny < 0) ny = 0;
  client->ch_ready[CMAD_CH_NY] = (ny == client->image_width) ? 1 : 0;

out:
  cmad_update_ready_locked(client);

  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for nz CA subscription state change events.
 */
static void
cmad_on_nz_state_change(struct event_handler_args args)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;
  dbr_long_t nz;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_NZ] = 0;
    goto out;
  }

  nz = *((dbr_long_t *)args.dbr);
  if (nz < 0) nz = 0;
  client->ch_ready[CMAD_CH_NZ] = (nz == client->image_height) ? 1 : 0;

out:
  cmad_update_ready_locked(client);

  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for color-mode CA subscription state change events.
 */
static void
cmad_on_color_mode_state_change(struct event_handler_args args)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_COLOR_MODE] = 0;
    goto out;
  }

  client->ch_ready[CMAD_CH_COLOR_MODE] =
      (*((dbr_enum_t *)args.dbr) == CMAD_AD_COLOR_MODE_RGB1) ? 1 : 0;

out:
  cmad_update_ready_locked(client);

  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for image-ID CA subscription state change events.
 */
static void
cmad_on_image_id_state_change(struct event_handler_args args)
{
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;
  int status;

  client = (struct cmad_client *)ca_puser(args.chid);

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_IMAGE_ID] = 0;
    cmad_update_ready_locked(client);
    goto out;
  } else {
    client->ch_ready[CMAD_CH_IMAGE_ID] = 1;
    cmad_update_ready_locked(client);
  }

  if (client->connected && client->ready_except_image) {
    status = ca_array_get_callback(DBR_CHAR, client->image_size,
        client->ch_image, cmad_on_image_get,
        cmad_encode_sequence_number(client->image_seq));
    if (status != ECA_NORMAL) goto out;
    status = ca_flush_io();
    if (status != ECA_NORMAL) goto out;
  }

out:
  epicsMutexUnlock(client->mutex_id);
}

/*
 * Callback function for image CA get.  Copy the image from the callback to the
 * client's image storage, and convert it into the native CameraMan in-memory
 * format.
 */
static void
cmad_on_image_get(struct event_handler_args args)
{
  ptrdiff_t seq_num;
  struct cmad_client *client;
  epicsMutexLockStatus mutex_status;
  int seq_signal_needed;

  seq_num = cmad_decode_sequence_number(args.usr);
  client = (struct cmad_client *)ca_puser(args.chid);
  seq_signal_needed = 0;

  mutex_status = epicsMutexLock(client->mutex_id);
  if (mutex_status != epicsMutexLockOK) return;

  if (args.status != ECA_NORMAL) {
    client->ch_ready[CMAD_CH_IMAGE] = 0;
    goto out;
  }

  if (args.count != client->image_size) {
    client->ch_ready[CMAD_CH_IMAGE] = 0;
    goto out;
  }

  (void)memcpy(client->image, args.dbr, client->image_size_in_bytes);
  cmad_rgb24_to_bgr24(client->image, client->image_size);
  if (cmad_is_sequence_number_valid(client->image_seq, seq_num)) {
    seq_signal_needed = 1;
  }

  client->ch_ready[CMAD_CH_IMAGE] = 1;

out:
  cmad_update_ready_locked(client);

  if (seq_signal_needed) epicsEventSignal(client->image_seq_eid);

  epicsMutexUnlock(client->mutex_id);
}
